/* Author: Venkataramana Badarla, email: badarla.venkataramana@nuim.ie
 * Platform: This module works on Linux 2.6.24.7 and written using click router modular architecture ver 1.7.0.
 * Purpose: Basically to identify flows and provide per-flow queueing.  
 *
 * For every transmission request from the driver, it schedules packets using round-robin scheduling.
 * However, scheduling is not restricted to just round-robin scheduler. It can be replaced simply with any other scheduler of interest. 
 * All control packets (like ARP handshaking packets) are stored in a separate queue and, given highest priority.
 * With every packet released to the driver, it appends (in TOS filed of IP header) the information of number of active flows.
 * Using this info, the EDCA parameters of IEEE 802.11e are changed in MadWifi driver.
 *
*/

#include <click/config.h>
#include "mmschd.hh"
#include <clicknet/ip.h>
#include <clicknet/tcp.h>
#include <clicknet/udp.h>
#include <clicknet/ether.h>
#include <click/sync.hh>
#include <click/error.hh>
#include <click/router.hh>
#include <click/confparse.hh>

CLICK_DECLS

#define STORE_CONTROL_PACKET() { \
  if (_ctlq._tail) \
         _ctlq._tail->set_next (p); \
  else \
         _ctlq._head = p; \
   _ctlq._tail = p; \
   p->set_next (0); \
  _ctlq._numpkts++; \
}

#define STORE_DATA_PACKET() { \
  if (_datq._tail) \
         _datq._tail->set_next (p); \
  else \
         _datq._head = p; \
   _datq._tail = p; \
   p->set_next (0); \
  _datq._numpkts++; \
}

# define DEAL_DATA_QUEUE() { \
 if(_datq._numpkts) { \
 	p = _datq._head; \
	if (!(_datq._head = p->next ())) \
              _datq._tail = 0; \
        --_datq._numpkts; \
   return p; \
} 

MMScheduler::MMScheduler () : _size_click_ip (0), _size_click_ether (0), _size_eth_ip(0), _lock(0), _flow_mgmt_timer(static_flow_mgmt_hook, this), _max_idle_time(500) // 500 milli seconds
{			
}

MMScheduler::~MMScheduler ()
{

}

int
MMScheduler::configure (Vector < String > &conf, ErrorHandler * errh)
{
  String channel;
  if (cp_va_kparse (conf, this, errh, "PHYRATE", 0, cpUnsigned, &_phy_rate, cpEnd) < 0)
    return -1;

  _errh = router()->chatter_channel (channel);
  _flow_mgmt_timer.initialize (this);
  _flow_mgmt_timer.schedule_after_msec (1000);

  return 0;
}


int
MMScheduler::initialize (ErrorHandler *errh)	
{
  _size_click_ip = sizeof (click_ip);
  _size_click_ether = sizeof (click_ether);
  _size_eth_ip = _size_click_ether + _size_click_ip;
   uint8_t rt[] = {6, 9, 12, 18, 24, 36, 48, 54};
   uint8_t rtval[] = {0x00, 0x20, 0x40, 0x60, 0x80, 0xA0, 0xC0, 0xE0};
   for(int i = 0; i < 8; i++) {
	_rates[rt[i]] = rtval[i];
   }

  _lock = new Spinlock();
  if (!_lock) {
    _errh->debug("cannot create spinlock.");
    return -1;
  }

  return 0;
}
void
MMScheduler::cleanup(CleanupStage)
{
  delete _lock;
  _lock = 0;
}

/* All the packets are received here */

void
MMScheduler::push (int port, Packet * p)
{
//  _lock->acquire();
   // if ip packet store it in appropriate queue 
  if (p->has_network_header() && port == 0) {
       append_packet (p);
// STORE_DATA_PACKET();	// append data packet
  }
  else {
	STORE_CONTROL_PACKET();	// append control packet
  }
//  _lock->release();
}

/* find flow control block, if it does exists create the one and return fcb */

FCB*
MMScheduler::find(IPAddress sip, uint16_t sport, IPAddress dip, uint16_t dport)
{
  FCB *fcb;
  IPFlowID q(sip, sport, dip, dport);

  if (fcb = _flows.get_pointer(q))
    return fcb;

   fcb = &_flows[q];
   fcb->srcip = sip;
   fcb->sport = ntohs(sport);
   fcb->dstip = dip;
   fcb->dport = ntohs(dport);
  return fcb;
} 
/* Once a packet is received append the packet in appropriate queue */
void MMScheduler::append_packet(Packet *p)
{
	click_ip *ip = (click_ip *) (p->data() + _size_click_ether); //p->ip_header();
        click_tcp *tcp; 
        click_udp *udp; 
	uint16_t sport;
	uint16_t dport;
	uint8_t index=0;

	IPAddress srcip = IPAddress (ip->ip_src);
	IPAddress dstip = IPAddress (ip->ip_dst);

        if(ip->ip_p == IP_PROTO_TCP) {
		click_tcp *tcp = (click_tcp *) (p->data() + _size_eth_ip); //p->ip_header();
	 	sport =  tcp->th_sport;
         	dport =  tcp->th_dport;
  		int datalen = p->length() - (ip->ip_hl << 2) + (tcp->th_off << 2);
		assert (datalen >= 0);
//		if (datalen == 0)	// tcp ack packet.. no need to go thru scheduler
//			goto store_ctlp;

        } else if(ip->ip_p == IP_PROTO_UDP){
		click_udp *udp = (click_udp *) (p->data() + _size_eth_ip); //p->ip_header();
                sport = udp->uh_sport;
                dport = udp->uh_dport;
        } else	// for ICMP etc packets
		goto store_ctlp;

 	// get or create corresponding Flow Control Block
    	FCB *fcb =  find(srcip,sport,dstip,dport);
 	// insert number of flows info in ip_tos field to adjust txop in madwifi
	if (_flows.size()) {
  	   ip->ip_tos = _flows.size();
	   ip->ip_tos = ip->ip_tos | _rates[_phy_rate];
	   // click_chatter("tos in mmschd %d fs: %d %d ",ntohs(ip->ip_tos), _flows.size(),ip->ip_tos);
	   ip->ip_sum = 0;
  	   ip->ip_sum = click_in_cksum ((const unsigned char *) ip, _size_click_ip);
	}
	fcb->_last_activity_time = Timestamp::now();
	if (fcb->_tail)
    		fcb->_tail->set_next (p);
  	else
    		fcb->_head = p;

  	fcb->_tail = p;
  	p->set_next (0);
	fcb->_numpkts++;
	return; 

store_ctlp:
   STORE_CONTROL_PACKET();
} 

/* Packets are released to the driver upon pull requests */ 
Packet *
MMScheduler::pull(int port)
{
  Packet *p = NULL;
// _lock->acquire();
  p=schd_packet();
// _lock->release();
	return p;
}

Packet *
MMScheduler::schd_packet() 
{
 
 FCB *fcb; 
 Packet *p = NULL;

 if(_ctlq._numpkts) { 
 	p = _ctlq._head;
	if (!(_ctlq._head = p->next ()))
              _ctlq._tail = 0; 
        --_ctlq._numpkts;
   return p;
 } 


// DEAL_DATA_QUEUE()

  if (!_flows.size())
 	return 0;

 if (!_next)
    _next = _flows.begin().key();
    FIter fj = _flows.find(_next);
    for (FIter fi = _flows.begin (); fi != _flows.end (); fi++) {
	fcb = &_flows[fj.key()];
	fj++;
	if (fj == _flows.end())
    	 	fj = _flows.begin();
	if (p = fcb->_head) {
		_next = fj.key();	
    		if (!(fcb->_head = p->next ()))
            		fcb->_tail = 0;
    		  --fcb->_numpkts;
		fcb->_last_activity_time = Timestamp::now();
		return p;
	}
   }
   return 0;
} 

enum { H_STATS };
void
MMScheduler::add_handlers ()
{
  add_read_handler ("stats", read_handler, H_STATS);
  add_write_handler ("maxidletime", write_handler, (void *) 0);
}


int
MMScheduler::write_handler (const String & s, Element * e, void *vparam, ErrorHandler * errh)
{
  MMScheduler *mw = static_cast < MMScheduler * >(e);
  switch ((intptr_t) vparam)
    {
    case 0:
      {
        int maxidletime;
        if (!cp_integer (s, &maxidletime))
          return errh->error ("maxidletime parameter must be integer");
        mw->_max_idle_time = maxidletime;
        break;
      }
    }
  return 0;
}



String MMScheduler::read_handler (Element * e, void *thunk)
{
  MMScheduler * mw = static_cast < MMScheduler * >(e);
  switch (reinterpret_cast < intptr_t > (thunk))
    {
    case H_STATS:
      {
        StringAccum sa;
        sa << "# of flows " << (0xFF & mw->_flows.size());
    	for (FIter fi = mw->_flows.begin (); fi != mw->_flows.end (); fi++) {
		FCB *fb = mw->_flows.get_pointer(fi.key());
		sa << "\nsip/sport " << fb->srcip << "/"<<fb->sport << " dip/dport " << fb->dstip <<"/"<<fb->dport;
	}	
        return sa.take_string ();
      }
    default:
      return String ();
    }
}
void
MMScheduler::flow_mgmt_hook()
{
 Timestamp time_now = Timestamp::now ();
// _lock->acquire();
 for (FIter fi = _flows.begin (); fi != _flows.end (); fi++) {
	FCB *fb = _flows.get_pointer(fi.key());
	if (!fb->_numpkts && fb->_last_activity_time && (time_now - fb->_last_activity_time).msecval() > _max_idle_time) {
	if (_next == fi.key())
	    _next = IPFlowID();
	_flows.erase(fi.key());
     }
 }
 _flow_mgmt_timer.schedule_after_msec(1000);
// _lock->release();
}


CLICK_ENDDECLS
EXPORT_ELEMENT (MMScheduler)
